%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Son Tran's translator program - mstr.pl - Jul 28, 2003               %%
%%                                                                      %%
%% input:  a Prolog program representing a domain                       %%
%%         description in Language B with sensing                       %%
%%         actions and multivalued fluents                              %%
%% output: a domain description and a domain problem                    %%
%%	   which will be used as input to Smodels                       %%
%%                                                                      %%
%% The b2s translator is currently capable of handling                  %%
%% the following features:                                              %%
%%	                                                                %%
%% a. Dynamic Causal Laws:                                              %%
%%    causes(A,L,P) iff action A causes fluent literal L                %%
%%	  to be true if precondition P hold. A fluent literal           %%
%%        is of either the form f = v or its negration f != v.          %%
%%        They are expressed in our language as eq(f,v) and neq(f,v)    %%
%%        respectively. Precondition P is a conjunction of such         %%
%%        fluent literals.                                              %%
%%                                                                      %%
%% b. Static Causal Laws:                                               %%
%%    caused(L,G) iff set of fluent literals G causes fluent            %%
%%	   literal L to be true.                                        %%
%%                                                                      %%
%% c. Executability conditions:                                         %%
%%    executable(A,G) - A can be executed iff set of fluent             %%
%%         literals G holds.                                            %%
%%                                                                      %%
%% d. Sensing actions:                                                  %%
%%    determines(A,F) - Action A determines the value of fluent F.      %%
%%                                                                      %%
%% e. Initial state:                                                    %%
%%    initially(O) iff set of fluent literals O holds in the            %%
%          initial state.                                               %%
%%                                                                      %%
%% f. Goal state:                                                       %%
%%    goal(G) iff set of fluent literals G holds in the final state.    %%
%%                                                                      %%
%% Notes:                                                               %%
%%    - abnormality rules were removed.                                 %%
%%    - branches do not occur if the sensing value is known to be       %%
%%          different than their values.                                %%
%%    - be careful with the statical causal laws. You must ensure that  %%
%%      they are alway satisfied in every situation.                    %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(library(lists)).
:- use_module(library(system)).

:- dynamic 
      noninertial/1,
      caused/2,
      causes/3,
      determines/2,
      initially/1,
      sgoal/1.

unknown_predicate_handler(_,_,fail).

% equality and unequality
eq(X,X).
neq(X,Y):- \+ eq(X,Y).

% main
pr_generals(Id) :-
    nl,
    write('Writing headings...'),
    pr_headings(Id),
    write('done.'), nl,
    nl(Id),
    write('Writing fluents and domains...'), 
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    write(Id, '% Fluents and domains'),
    nl(Id),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    findall(X, fluent(X), LF),
    nl(Id),
    write(Id,'% fluents'), nl(Id),
    pr_fluents(LF,Id),
    findall(NF,noninertial(NF),LNF),
    write(Id,'% non-inertial fluents'), nl(Id),
    pr_noninertials(LNF,Id),
    write(Id,'% domains'), nl(Id),
    findall((F,V),domain(F,V),LD),
    pr_domains(LD,Id),
    write(Id,'% domain restriction rules'), nl(Id),
    pr_domain_restriction_rules(LF,Id),
    write('done.'), nl,
    write('Writing actions...'),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    write(Id, '% Actions'),
    nl(Id),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    findall(Y,action(Y),LA),
    pr_actions(LA,Id),
    findall((SA,SF),determines(SA,SF),LSA),
    pr_sens_actions(LSA,Id),
    write('done.'), nl,
    write('Writing executability conditions...'),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    write(Id, '% Executablility Conditions'),
    nl(Id),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    findall((EXA,LEXCOND),executable(EXA,LEXCOND),LEXACOND),
    pr_ex_conditions(LEXACOND,Id),
    write('done.'), nl,
    write('Writing effects....'),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    write(Id, '% Effects'),
    nl(Id),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    findall((EFAct,EFFluent,EFConds),causes(EFAct,EFFluent,EFConds),EFEffects),
    pr_effects(EFEffects,Id),
    pr_sens_effects(LSA,Id),
    write('done.'), nl,
    write('Writing statical causal laws....'),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    write(Id, '% Statical laws'),
    nl(Id),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    pr_statical_laws(Id),
    nl(Id),
    write('done.'), nl,
    write('Writing initial situation...'),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    write(Id, '% Initial Situation'),
    nl(Id),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    findall(Init,initially(Init),LInits),
    pr_inits(LInits,Id),
    write('done.'), nl,
    write('Writing goal specification...'),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    write(Id, '% Goal specification'),
    nl(Id),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    findall(Goal,sgoal(Goal),LGoals),
    pr_goals(LGoals,Id),
    nl(Id),
    write('done.'), nl,
    write('Writing independent rules...'),
    pr_independent_rules(Id),
    write('done.'), nl,
    write('Writing hide section...'),
    pr_hide_section(Id),
    write('done.'), nl.

% print headings
pr_headings(Id) :- write(Id, 'time(1..length).'), nl(Id),
    write(Id,'path(1..level).'), nl(Id).

% print fluents
pr_fluents([],Id) :- nl(Id).

pr_fluents([X|Y],Id) :-
    write(Id,fluent(X)), write(Id,'.'), nl(Id),
    pr_fluents(Y,Id).

% print noninertial fluents
pr_noninertials([],Id) :- nl(Id).

pr_noninertials([X|Y],Id) :-
    write(Id,noninertial(X)), write(Id,'.'), nl(Id),
    pr_noninertials(Y,Id).

% print domains
pr_domains([],Id) :- nl(Id).

pr_domains([(F,V)|T],Id) :- write(Id,'domain('), write(Id,F), 
    write(Id,','), write(Id,V), write(Id,').'), nl(Id), pr_domains(T,Id).

% print domain restriction rules
pr_domain_restriction_rules([],Id) :- nl(Id).
pr_domain_restriction_rules([F|T],Id) :- findall(V, domain(F,V), LV),
    pr_domain_restriction_rule(F,LV,LV,Id), nl(Id), pr_domain_restriction_rules(T,Id).  

pr_domain_restriction_rule(_,[],_,_).
pr_domain_restriction_rule(F,[V|T],LV,Id) :- 
     pr_atom(eq(F,V),'T,P',Id), write(Id,' :- time(T), path(P)'),
     select(V,LV,LV1),
     pr_neg_list(F,LV1,Id), nl(Id), pr_domain_restriction_rule(F,T,LV,Id).

pr_neg_list(_,[],Id) :- write(Id,'.').
pr_neg_list(F,[V|LV],Id) :- write(Id,','), nl(Id), write(Id,'     '), 
    pr_atom(neq(F,V),'T,P',Id), pr_neg_list(F,LV,Id).

% print actions
pr_actions([],Id) :- nl(Id).

pr_actions([X|Y],Id) :-
    write(Id,action(X)),write(Id,'.'), nl(Id),
    pr_actions(Y,Id).

% print sensing actions
pr_sens_actions([],Id) :- nl(Id).

pr_sens_actions([(A,F)|T],Id) :-
    write(Id,'determines('), write(Id,A), write(Id,','),
    write(Id,F), write(Id,').'), nl(Id), 
    write(Id,'sense('), write(Id,F), write(Id,').'), nl(Id),
    pr_sens_actions(T,Id).

% print actions' executability conditions
pr_ex_conditions([],Id) :- nl(Id).
pr_ex_conditions([(A,LC)|T],Id) :-
    write(Id,'possible('), write(Id,A), write(Id,',T,P) :- time(T), path(P)'),
    pr_conditions(LC,'',Id), write(Id, '.'), nl(Id), pr_ex_conditions(T,Id).

pr_conditions([],_,_). 
pr_conditions([E|T],Prefix,Id) :-  pr_condition(E,Prefix,Id), pr_conditions(T,Prefix,Id).

pr_condition(FV,Prefix,Id) :- write(Id,','), nl(Id), write(Id,'    '), write(Id,Prefix), 
    pr_atom(FV,'T,P',Id).

% print effects
pr_effects([],Id) :- nl(Id).
pr_effects([H|T],Id) :- pr_effect(H,Id), nl(Id), nl(Id), pr_effects(T,Id).

pr_effect((A,FV,[]),Id) :- pr_atom(FV,'T+1,P',Id), write(Id,' :- time(T), path(P), '), 
    nl(Id),
    write(Id,'    occ('),
    write(Id,A), write(Id,',T,P).').

pr_effect((A,FV,LC),Id) :- pr_atom(FV,'T+1,P',Id), write(Id,' :- time(T), path(P)'), 
    pr_conditions(LC,'',Id), write(Id, ','), nl(Id),
    write(Id,'    occ('),
    write(Id,A), write(Id,',T,P).').   

% prin sensing effects
pr_sens_effects([],Id) :- nl(Id).
pr_sens_effects([(A,F)|T],Id) :- pr_sens_effect(A,F,Id), pr_sens_effects(T,Id).

pr_sens_effect(A,F,Id) :- findall((F,V),domain(F,V),LV), pr_branches(A,LV,Id).

pr_branches(_,[],Id) :- nl(Id).
pr_branches(A,[(F,V)|T],Id) :- write(Id,'1{branch('), write(Id,F), write(Id, ','),
    write(Id,V), write(Id,',T,P,P1):new_branch(P,P1)}1 :- time(T), path(P),'), nl(Id),
    write(Id,'    not '), pr_atom(neq(F,V),'T,P',Id), write(Id,','), nl(Id),
    write(Id,'    occ('), write(Id,A), write(Id,',T,P).'), nl(Id),
    nl(Id), pr_branches(A,T,Id).

% print statical laws
pr_atom(eq(F,V), Postfix,Id) :- write(Id,'holds('), write(Id,F), write(Id,','),
    write(Id,V), write(Id,','), write(Id,Postfix), write(Id,')').

pr_atom(neq(F,V),Postfix,Id) :- write(Id,'neg_holds('), write(Id,F), write(Id,','),
    write(Id,V), write(Id,','), write(Id,Postfix), write(Id,')').

% Collecting static causal laws
% and translating them into rules suitable for SMODELS
pr_statical_laws(Id):-
    findall((X,Y),caused(X,Y),L),
    nl(Id), 
    pr_static_causal(L,Id).

pr_static_causal([],Id):- nl(Id).

pr_static_causal([(E,L)|T], Id):-
    pr_atom(E,'T,P',Id), write(Id,' :-  time(T), path(P)'),
    pr_conditions(L,'',Id), write(Id,'.'), nl(Id), pr_static_causal(T,Id).

% print initial situation
pr_inits([],Id) :- nl(Id).
pr_inits([H|T],Id) :- pr_init(H,Id), pr_inits(T,Id).

pr_init([],_).
pr_init([H|T],Id) :- pr_atom(H,'1,1',Id), write(Id,'.'), nl(Id), pr_init(T,Id).

% print goals
pr_goals(LGoals,Id) :- write(Id,'used(1,1).'), nl(Id),
    write(Id,'used(T+1,P) :- time(T), path(P), used(T,P).'),
    nl(Id), nl(Id),
    write(Id,'goal(T,P) :- time(T), path(P)'),
    pr_sgoals(LGoals,Id),
    nl(Id),
    write(Id,'ggoal(P) :- path(P), time(T), used(T,P), goal(length,P).'),
    nl(Id), nl(Id),
    write(Id,':- path(P), used(length,P), not ggoal(P).'), nl(Id).

pr_sgoals([],Id) :- write(Id,'.'), nl(Id) .
pr_sgoals([H|T],Id) :- pr_conditions(H,'',Id), pr_sgoals(T,Id).

% domain independent rules
pr_independent_rules(Id) :-
    write(Id,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    write(Id,'% Branch Restriction Rules'), nl(Id),
    write(Id,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    nl(Id),
    write(Id,':- time(T), path(P1), path(P2),'), nl(Id),
    write(Id,'    path(L1), P1 <= P2, L1 <= P2,'), nl(Id),
    write(Id,'    sense(F), sense(G), neq(F,G),'), nl(Id),
    write(Id,'    domain(F,V), domain(G,V1),'), nl(Id),
    write(Id,'    branch(F,V,T,P1,P2),'), nl(Id),
    write(Id,'    branch(G,V,T,L1,P2).'), nl(Id),
    nl(Id),
    write(Id,':- time(T), T > 1,'), nl(Id),
    write(Id,'    path(P1), path(P2), P1 < P2,'), nl(Id),
    write(Id,'    sense(F), domain(F,V),'), nl(Id),
    write(Id,'    branch(F,V,T,P1,P2), used(T,P2).'), nl(Id),
    nl(Id),
    write(Id,'new_branch(P,P1) :- path(P), path(P1),P <= P1.'), nl(Id),
    nl(Id),
    write(Id,'branch(F,V,T,P,P1) :- sens(F), domain(F,V), time(T), path(P),'), nl(Id),
    write(Id,'     path(P1), P <= P1, path(P2), P <= P2, branch(F,V,T,P,P2).'), nl(Id),
    nl(Id),
    write(Id,'% mark one node as used if there is a branch at it'), nl(Id),
    write(Id,'used(T+1,P1):- path(P), time(T), path(P1), P <= P1,'), nl(Id),
    write(Id,'    sense(F), domain(F,V), branch(F,V,T,P,P1).'), nl(Id),
    nl(Id),
    write(Id,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'), nl(Id),
    write(Id,'% Inertial rules'), nl(Id),
    write(Id,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'), nl(Id),
    write(Id,'% one fluent has only value at any point of time.'), nl(Id),
    write(Id,':- time(T), path(P), domain(F,V),'), nl(Id),
    write(Id,'    holds(F,V,T,P), neg_holds(F,V,T,P).'), nl(Id),
    nl(Id),
    write(Id,'neg_holds(F,V,T,P) :- time(T), path(P), fluent(F), domain(F,V),'), nl(Id),
    write(Id,'    domain(F,V1), neq(V,V1), holds(F,V1,T,P).'), nl(Id),
    nl(Id),
    write(Id,'holds(F,V,T+1,P) :-'), nl(Id),
    write(Id,'    time(T), path(P),'), nl(Id),
    write(Id,'    fluent(F), domain(F,V),'), nl(Id),
    write(Id,'    not noninertial(F),'), nl(Id),
    write(Id,'    holds(F,V,T,P),'), nl(Id),
    write(Id,'    not neg_holds(F,V,T+1,P).'), nl(Id),
    nl(Id),
    write(Id,'neg_holds(F,V,T+1,P) :-'), nl(Id),
    write(Id,'    time(T), path(P),'), nl(Id),
    write(Id,'    fluent(F), domain(F,V),'), nl(Id),
    write(Id,'    not noninertial(F),'), nl(Id),
    write(Id,'    neg_holds(F,V,T,P),'), nl(Id),
    write(Id,'    not holds(F,V,T+1,P).'), nl(Id),
    nl(Id),
    write(Id,'% for sensing actions:'), nl(Id),
    write(Id,'holds(F,V,T+1,P1):-'), nl(Id),
    write(Id,'    time(T), path(P1), path(P), P <= P1,'), nl(Id),
    write(Id,'    fluent(F), domain(F,V), '), nl(Id),
    write(Id,'    sense(G), domain(G,V1),'), nl(Id),
    write(Id,'    neq(F,G),'), nl(Id),
    write(Id,'    branch(G,V1,T,P,P1),'), nl(Id),
    write(Id,'    holds(F,V,T,P).'), nl(Id),
    nl(Id),
    write(Id,'neg_holds(F,V,T+1,P1):-'), nl(Id),
    write(Id,'    time(T), path(P1), path(P), P <= P1,'), nl(Id),
    write(Id,'    fluent(F), domain(F,V),'), nl(Id),
    write(Id,'    sense(G), domain(G,V1),'), nl(Id),
    write(Id,'    neq(F,G),'), nl(Id),
    write(Id,'    branch(G,V1,T,P,P1),'), nl(Id),
    write(Id,'    neg_holds(F,V,T,P).'), nl(Id),
    nl(Id),
    write(Id,'% for sensing actions'), nl(Id),
    write(Id,'holds(F,V,T+1,P1) :- domain(F,V), time(T), path(P),'), nl(Id),
    write(Id,'    path(P1), P <= P1, sense(F), branch(F,V,T,P,P1).'), nl(Id),
    nl(Id),       
    write(Id,'% possibly holds'), nl(Id),
    write(Id,'poss_holds(F,V,T,P):-'), nl(Id),
    write(Id,'    fluent(F), domain(F,V),'), nl(Id),
    write(Id,'    time(T), path(P), not neg_holds(F,V,T,P).'), nl(Id),
    nl(Id),
    write(Id,'% possibly neg holds'), nl(Id),
    write(Id,'poss_neg_holds(F,V,T,P):-'), nl(Id),
    write(Id,'    fluent(F), domain(F,V),'), nl(Id),
    write(Id,'    time(T), path(P), not holds(F,V,T,P).'), nl(Id),
    nl(Id),
    write(Id,'% a fluent is known if it is true or false'), nl(Id),
    write(Id,'known(F,T,P) :-'), nl(Id),
    write(Id,'    path(P), time(T), fluent(F), domain(F,V), holds(F,V,T,P).'), nl(Id),
    nl(Id),
    write(Id,'% no sensing action needed if the fluent is known'), nl(Id),
    write(Id,':- action(A), fluent(F), time(T), path(P),'), nl(Id),
    write(Id,'    determines(A,F), occ(A,T,P), known(F,T,P).'), nl(Id),
    nl(Id),
    write(Id,'% at least one action must occur at (T,P) if the goal'), nl(Id),
    write(Id,'% has not been reached at (T,P)'), nl(Id),
    write(Id,'1{occ(A,T,P) : action(A)}1:-'), nl(Id),
    write(Id,'    time(T),'), nl(Id),
    write(Id,'    path(P),'), nl(Id),
    write(Id,'    used(T,P),'), nl(Id),
    write(Id,'    not goal(T,P).'), nl(Id),
    nl(Id),
    write(Id,'% an action cannot occur if it is not executable'), nl(Id), 
    write(Id,':-  time(T),  path(P),'), nl(Id),
    write(Id,'    used(T,P), action(A),'), nl(Id),
    write(Id,'    occ(A,T,P), not possible(A,T,P).'), nl(Id).

pr_hide_section(Id) :-
    nl(Id),
    write(Id,'% the following are hidden'), nl(Id),
    write(Id,'hide domain(X,V).'), nl(Id),
    write(Id,'hide fluent(F).'), nl(Id),
    write(Id,'hide noninertial(F).'), nl(Id),
    write(Id,'hide action(A).'), nl(Id),
    write(Id,'hide time(T).'), nl(Id),
    write(Id,'hide path(P).'), nl(Id),
    write(Id,'hide determines(A,F).'), nl(Id),
    write(Id,'hide boolean(B).'), nl(Id),
    write(Id,'hide possible(A,T,P).'), nl(Id),
    write(Id,'hide known(F,T,P).'), nl(Id),
    write(Id,'hide unknown(F,T,P).'), nl(Id),
    write(Id,'hide holds(F,V,T,P).'), nl(Id),
    write(Id,'hide used(T,P).'), nl(Id),
    write(Id,'hide poss_holds(F,V,T,P).'), nl(Id),
    write(Id,'hide goal(T,P).'), nl(Id),
    write(Id,'hide ggoal(P).'), nl(Id),
    write(Id,'hide sense(F).'), nl(Id),
    write(Id,'hide neg_holds(F,V,T,P).'), nl(Id),
    write(Id,'hide poss_neg_holds(F,V,T,P).'), nl(Id),
    write(Id,'hide new_branch(P,P1).'), nl(Id).

gen:-
    write(' Enter domain description file name: '),
    read(S),
    name(S,Chars),
    name(smo,Chars1),
    append(Chars,[46|Chars1],Chars2),
    name(S1,Chars2),
    absolute_file_name(S, Source), %%% get the source file name
    consult(Source),
    absolute_file_name(S1, FileDescr), %%% get the domain output file name
    write('Generating .... '),
    open(FileDescr, write , Id),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    write(Id, '% Domain Description, generated by the translator'),
    nl(Id),
    write(Id, '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
    nl(Id),
    pr_generals(Id),
    close(Id).
